#include <stdlib.h>
#include <unistd.h>
#include <signal.h>
#include <iostream.h>
#include <string.h> 
#include <errno.h>

#include "debug.hh"

#include "list.hh"
#include "buffer.hh"

#ifdef __sun__
#include "b.hh"
#endif

/************************************************************************
**                             RTP_BUFFER                              **
************************************************************************/


Buffer::Buffer(unsigned int _size)
: FlatMemory(_size)
{
  PDEBUG("Making a Buffer");
  if(address == NULL)
    {
      cerr << "BUFFER: no more memory for buffers\n";
      exit(-1);
    }
  status = BUFFER_STATUS_FREE;
  PDEBUG2("Buffer memory created at address ",(void*)address);
  PDEBUG2("Buffer created at address ",(void*)this);
}

Buffer::~Buffer()
{
  PDEBUG("Delete a Buffer");
  if(status == BUFFER_STATUS_WAIT)
    cerr << "BUFFER: warning deleting buffer in WAIT status\n";
}


void Buffer::RemUse()
{
  if(use_count == 2)
    status = BUFFER_STATUS_FREE;

  FlatMemory::RemUse();
}

/*
unsigned int Buffer::Read(void* pData,unsigned int length,unsigned int offset)
{
  return FlatMemory::Read(pData,length,offset)
}
*/

/*

{
  if(offset >= size)
    return 0;

  if((length + offset) > size)
    length = size - offset;
  
  bcopy(((char *) address) + offset,(char*) pData,length);
  return length;
}
  
unsigned int Buffer::Write(void* pData,unsigned int length,unsigned int offset)
{
  if(offset >= size)
    return 0;

  if((length + offset) > size)
    length = size - offset;
  
  if((length + offset) > data_length)
    data_length = length + offset;

  bcopy((char *) pData,((char*) address) + offset,length);  
  return length;
}
*/

int Buffer::Status()
{
  return status;
}

void Buffer::FreeBuffer()
{
  status = BUFFER_STATUS_FREE;
  //cout << "BUFFER: Set buffer to free\n";
}

void Buffer::MarkBuffer()
{
  status = BUFFER_STATUS_WAIT;
}


void* Buffer::BufferAddress()
{
  return address;
}



/************************************************************************
**                           RTP_BUFFERPOOL                            **
************************************************************************/



void BufferPool::MakeFreeList()
{
/*
  for(int i=1;i <= free_list.Length();i++)
    if(free_list[i].Status() == BUFFER_STATUS_WAIT)
      {
	PDEBUG("Removing a Buffer to free list");
	free_list.Remove(i);
      }
*/
  
  for(int i=1;i <= list.Length();i++)
    {
      //cerr << "buffer #" << i << " status:\t" << list[i].Status() << '\n';
      if(list[i].Status() == BUFFER_STATUS_FREE)
	{
	  PDEBUG("Adding a Buffer to free list");
	  free_list.AddTail(&list[i]);
	}
    }
}
  

BufferPool::BufferPool(int number, int size)
{
  if(number > DEFAULT_BUFFER_POOL_MAX_LENGTH*5)
    cout << "warning: you are allocating " << number << " buffer\n";

  Buffer* pBuffer;

  length = number;
  buffer_size = size;
  for(int i=0;i < number;i++)
    {
      pBuffer = new Buffer(size);
      if(pBuffer == NULL)
	{
	  cerr << "no more memory for Buffer in BufferPool\n";
	  break;
	  //kill(SIGTERM,getppid());
	}
      PDEBUG("Adding a Buffer to buffer list");
      list.AddTail(pBuffer);
    }
  MakeFreeList();
}

BufferPool::~BufferPool()
{
  for(int i = 1;i <= list.Length();i++)
    delete &list[i];
}

Buffer* BufferPool::QueryBuffer()
{
  //static unsigned int bcnt = 0;
  Buffer* pBuffer = NULL;

  if(free_list.Length() == 0)
    {
      this->MakeFreeList();
      if(free_list.Length() == 0)
	{
	  cerr << "BUFFER_POOL: All the buffers are full; Warning!!\n";
	  if(buffer_size * list.Length() < (1048576 * 8))
	    {
	      for(int i=0;i < 50;i++)
		{
		  pBuffer = new Buffer(buffer_size);
		  if(pBuffer == NULL)
		    {
		      cerr << "no more memory for Buffer\n";
		      break;
		      //kill(SIGTERM,getppid());
		    }
		  PDEBUG("Adding a Buffer to buffer list");
		  list.AddTail(pBuffer);
		}
	      MakeFreeList();
	    }
	  else
	    {
	      cerr << "SORRY!! All the buffers are full, and i can't create another one.\n";
	      return NULL;
	    }	  
	}
    }
  Buffer* pbuffer = &free_list[1];
  free_list.RemHead();
  pbuffer->MarkBuffer();
  return pbuffer;
}


unsigned int BufferPool::QueryBufferVector(Buffer** bufv,unsigned int bufs)
{
  for(int i=0;i<bufs;i++)
    if((bufv[i] = this->QueryBuffer()) == NULL)
      {
	this->MakeFreeList();
	if((bufv[i] = this->QueryBuffer()) == NULL)
	  return i;
      }
    else
      bufv[i]->MarkBuffer();

  return bufs;
}

void BufferPool::ReleaseBufferVector(Buffer** bufv,unsigned int bufs)
{
  for(int i=0;i<bufs;i++)
    if(bufv[i] != NULL)
      bufv[i]->FreeBuffer();
}  

unsigned int BufferPool::BuffersSize()
{
  return buffer_size;
}





